def combine_dict(*d):
    """
    Returns a dict with all keys+values of all dict arguments.
    The first found value wins.
    
    This recurses if values are dicts.
    """
    res = {}
    keys = {}
    if len(d) <= 1:
        return d
    for kv in d:
        for k,v in kv.items():
            if k not in keys:
                keys[k] = []
            keys[k].append(v)
    for k,v in keys.items():
        if len(v) == 1:
            res[k] = v[0]
        elif not isinstance(v[0],Mapping):
            for vv in v[1:]:
                assert not isinstance(vv,Mapping)
            res[k] = v[0]
        else:
            res[k] = combine_dict(*v)
    return res

class attrdict(dict):
    """A dictionary which can be accessed via attributes, for convenience"""
    def __init__(self,*a,**k):
        super(attrdict,self).__init__(*a,**k)
        self._done = set()

    def __getattr__(self,a):
        if a.startswith('_'):
            return super(attrdict,self).__getattr__(a)
        try:
            return self[a]
        except KeyError:
            raise AttributeError(a)
    def __setattr__(self,a,b):
        if a.startswith("_"):
            super(attrdict,self).__setattr__(a,b)
        else:
            self[a]=b
    def __delattr__(self,a):
        del self[a]

import yaml
from yaml.representer import SafeRepresenter
SafeRepresenter.add_representer(attrdict, SafeRepresenter.represent_dict)


class PathShortener:
    """This class shortens path entries so that the initial components that
    are equal to the last-used path (or the original base) are skipped.

    It is illegal to path-shorten messages whose path does not start with
    the initial prefix.

    Example: The sequence

        a b
        a b c d
        a b c e f
        a b c e g h
        a b c i 
        a b j 

    is shortened to

        0
        0 c d
        1 e f
        2 g h
        1 i
        0 j

    where the initial number is the passed-in depth.

    Usage::

        >>> d = _PathShortener(['a','b'])
        >>> d({'path': 'a b c d'.split})
        {'depth':0, 'path':['c','d']}
        >>> d({'path': 'a b c e f'.split})
        {'depth':1, 'path':['e','f']}

    etc.

    Note that the dict is modified in-place.

    """
    def __init__(self, prefix):
        self.prefix = prefix
        self.depth = len(prefix)
        self.path = []

    def __call__(self, res):
        if res.path[:self.depth] != self.prefix:
            raise RuntimeError("Wrong prefix: has %s, want %s" % (repr(res.path), repr(self.prefix))

        p = res['path'][self.depth:]
        cdepth = min(len(p), len(self.path))
        for i in range(cdepth):
            if p[i] != self.path[i]:
                cdepth = i
                break
        self.path = p
        p = p[cdepth:]
        res['path'] = p
        res['depth'] = cdepth


class PathLongener:
    """This reverts the operation of a PathShortener.
    """
    def __init__(self, prefix):
        self.depth = len(prefix)
        self.path = prefix

    def __call__(self, res):
        p = res.pop('path', ())
        d = res.pop('depth', None)
        if d is not None:
            p = self.path[:self.depth+d] + p
        self.path = p
        res['path'] = p


class MsgReader:
    """Read a stream of messages from a file.
    
    Usage::

        with MsgReader("/tmp/msgs.pack") as f:
            for msg in f:
                process(msg)
    """

    def __init__(self, filename=None, stream=None):
        if (filename is None) == (stream is None):
            raise RuntimeError("You need to specify either filename or stream")
        from .codec import stream_unpacker
        self.fn = filename
        self.fd = stream
        self.unpack = stream_unpacker()

    def __enter__(self):
        if self.fd is None:
            self.fd = open(self.fn,"rb")
        return self

    def __exit__(self, *tb):
        if self.fn is not None:
            self.fd.close()

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            try:
                msg = next(self.unpacker)
            except StopIteration:
                pass
            else:
                return msg
            
            d = self.fd.read(4096)
            if d == b"":
                raise StopIteration
            unpacker.feed(d)

class MsgWriter:
    """Write a stream of messages from a file.
    
    Usage::

        with MsgWriter("/tmp/msgs.pack") as f:
            for msg in some_source_of_messages():
                f(msg)
    """

    def __init__(self, fn, buflen=65536):
        from .codec import stream_unpacker
        self.fn = fn
        self.buf = []
        self.buflen = buflen
        self.curlen = 0

    def __enter__(self):
        self.fd = open(fn,"rb")
        return self

    def __exit__(self, *tb):
        if self.buf:
            self.fd.write(self.buf)
        self.fd.close()

    def __call__(self, msg):
        buf = self.buf + packer(msg)
        if len(buf) >= self.buflen:
            pos = int(len(buf/self.buflen))
            self.fd.write(buf[:pos])
            buf = buf[pos:]
        self.buf = buf

        while True:
            try:
                msg = next(self.unpacker)
            except StopIteration:
                pass
            else:
                return msg
            
            d = self.fd.read(4096)
            if d == b"":
                raise StopIteration
            unpacker.feed(d)




from .util import attrdict

__all__ = [
        'PORT', 'CFG',
    ]

PORT = 27586  # 2000 + 100*ord('K') + ord('V')

CFG = attrdict(
        server=attrdict(
            host="localhost",
            port=PORT,
            ),
        serf=attrdict(
            host="localhost",
            port=7373,
            ),
        state=None, # path to load/save system state
    )



import trio

# We can just use 'async def test_*' to define async tests.
# This also uses a virtual clock fixture, so time passes quickly and
# predictably.
async def test_sleep_with_autojump_clock(autojump_clock):
    assert trio.current_time() == 0

    for i in range(10):
        print("Sleeping {} seconds".format(i))
        start_time = trio.current_time()
        await trio.sleep(i)
        end_time = trio.current_time()

        assert end_time - start_time == i



# Local server

import anyio
from anyio.exceptions import ClosedResourceError
from async_generator import asynccontextmanager
import msgpack
from aioserf import serf_client

from .model import Entry, NodeEvent, Node, Watcher, UpdateEvent
from .util import attrdict, PathShortener, PathLongener
from . import _version_tuple

import logging
logger = logging.getLogger(__name__)

_packer = msgpack.Packer(strict_types=False, use_bin_type=True).pack

class ClientError(ValueError):
    """report error to the client but don't dump a stack trace"""
    pass

class _LATER:
    pass

_client_nr = 0


def _multi_response(fn):
    async def wrapper(self, msg):
        seq = msg.seq
        await self.send({'seq':seq, 'state':'start'})
        try:
            await fn(self, msg)
        except BaseException as exc:
            raise # handled in the caller
        else:
            await self.send({'seq':seq, 'state':'end'})
    return wrapper

def _stream(fn):
    """A simple wrapper that calls the original function with each message"""
    async def wrapper(self, msg):
        seq = msg.seq
        q = self.inqueue[seq]
        path = msg.get('path', None)
        longer = PathLongener(path) if path is not None else lambda x:x
        n=0
        while True:
            msg = await q.get()
            if msg.get('state','') == 'end':
                await self.send({'seq':seq, 'count':n})
                del self.inqueue[seq]
                return
            longer(msg)
            res = await fn(self, msg)
            if isinstance(res, int):
                n += res
    return wrapper

class ServerClient:
    """Represent one listening client"""
    _nursery = None

    def __init__(self, server, root, stream):
        self.server = server
        self.root = root
        self.stream = stream
        self.seq = 0
        self.tasks = {}
        self._chop_path = 0

        global _client_nr
        _client_nr += 1
        self._client_nr = _client_nr
        logger.debug("CONNECT %d %s", self._client_nr, repr(stream))
    
    async def process(self, msg, stream=False):
        """
        Process an incoming message.
        """
        seq = msg.seq
        async with anyio.open_cancel_scope() as s:
            self.tasks[seq] = s
            try:
                if 'chain' in msg:
                    msg.chain = NodeEvent.unserialize(msg.chain)

                try:
                    fn = getattr(self, ('scmd_' if stream else 'cmd_') + str(msg.action))
                except AttributeError:
                    raise ClientError("Command not recognized: " + repr(msg.action))
                else:
                    res = await fn(msg)
                    if res is not None:
                        await self.send_result(seq, res)

            except BrokenPipeError as exc:
                logger.error("ERR%d: %s", self._client_nr, repr(exc))

            except Exception as exc:
                if not isinstance(exc, ClientError):
                    logger.exception("ERR%d: %s", self._client_nr, repr(msg))
                await self.send({'error': str(exc), 'seq': seq})

            finally:
                del self.tasks[seq]
        
    async def cmd_root(self, msg):
        """Change to a sub-tree.
        """
        entry = self.root.follow(*msg.path)
        self.root = entry
        self._chop_path += len(msg.path)
        return entry.serialize(chop_path=self._chop_path)

    async def cmd_get_value(self, msg):
        """Get a node's value.
        """
        try:
            entry = self.root.follow(*msg.path, create=False)
        except KeyError:
            entry = {'value': None}
        else:
            entry = entry.serialize(chop_path=-1, nchain=msg.get('nchain', 0))
        return entry

    async def cmd_set_value(self, msg):
        """Set a node's value.
        """
        entry = self.root.follow(*msg.path)
        send_prev = True
        if 'prev' in msg:
            if entry.data != msg.prev:
                raise ClientError("Data is %s" % (repr(entry.data),))
            send_prev = False
        if 'chain' in msg:
            if entry.chain is not None:
                if msg.chain is None:
                    raise ClientError("This entry already exists")
                if entry.chain != msg.chain:
                    raise ClientError("Chain is %s" % (repr(entry.chain),))
            elif msg.chain is not None:
                raise ClientError("This entry is new")
            send_prev = False
        res = {'changed': entry.data != msg.value}
        if send_prev:
            res['prev'] = entry.data

        async with self.server.next_event() as event:
            await entry.set_data(event, msg.value)

        nchain = msg.get('nchain', 1)
        if nchain > 0:
            res['chain'] = entry.chain.serialize(nchain=nchain)
        return res

    @_stream
    async def scmd_update(self, msg):
        """
        Apply a stored update.

        This streams the input.
        """
        msg = UpdateEvent.unserialize(self.root, msg)
        await msg.entry.apply(msg)

    async def cmd_update(self, msg):
        """
        Apply a stored update.

        You usually do this via a server command.
        """
        msg = UpdateEvent.unserialize(self.root, msg)
        res = await msg.entry.apply(msg)
        if res is None:
            return False
        else:
            return res.serialize(chop_path=self._chop_path)

    async def cmd_delete_value(self, msg):
        """Delete a node's value.
        Sub-nodes are not affected.
        """
        if 'value' in msg:
            raise ClientError("A deleted entry can't have a value")
        msg.value = None
        return await self.cmd_set_value(msg)

    @_multi_response
    async def cmd_get_tree(self, msg):
        try:
            entry = self.root.follow(*msg.path, create=False)
        except KeyError:
            await self.send_result(seq, {'value':None})
            return

        seq = msg.seq
        nchain = msg.get('nchain',0)
        shorter = PathShortener(entry.path)

        async def _get_values(entry, cdepth=0):
            if entry.data is not None:  # TODO: send if recently deleted
                res = entry.serialize(chop_path=self._chop_path, nchain=nchain)
                shorter(res)
                await self.send_result(seq, res)
            for v in list(entry.values()):
                await _get_values(v, cdepth=cdepth)

        await _get_values(entry)

    @_multi_response
    async def cmd_watch(self, msg):
        entry = self.root.follow(*msg.path, create=True)
        seq = msg.seq
        nchain = msg.get('nchain',0)

        async with Watcher(entry) as watcher:
            shorter = PathShortener(entry.path)

            async for msg in watcher:
                res = msg.entry.serialize(chop_path=self._chop_path, nchain=nchain)
                shorter(res)
                await self.send_result(seq, res)


    async def cmd_delete_tree(self, msg):
        """Delete a node's value.
        Sub-nodes are cleared (after their parent).
        """
        seq = msg.seq
        nchain = msg.get('nchain', 0)
        if nchain:
            await self.send({'seq':seq, 'state':'start'})
        ps = PathShortener(msg.path)

        try:
            entry = self.root.follow(*msg.path)
        except KeyError:
            return False

        async def _del(entry):
            res = 0
            if entry.data is not None:
                async with self.server.next_event() as event:
                    evt = await entry.set_data(event, None)
                    if nchain:
                        r = evt.serialize(chop_path=self._chop_path, nchain=nchain, with_old=True)
                        r['seq'] = seq
                        del r['new_value']  # always None
                        ps(r)
                        await self.send(r)
                res += 1
            for v in entry.values():
                res += await _del(v)
            return res
        res = await _del(entry)
        if nchain:
            await self.send({'seq':seq, 'state':'end'})
        else:
            return {'changed': res}

    async def cmd_stop(self, msg):
        try:
            t = self.tasks[msg.task]
        except KeyError:
            return False
        t.cancel()
        return True

    def send(self, msg):
        logger.debug("OUT%d: %s", self._client_nr, msg)
        return self.stream.send_all(_packer(msg))

    def send_result(self, seq, res):
        res['seq'] = seq
        return self.send(res)

    async def run(self):
        unpacker = msgpack.Unpacker(object_pairs_hook=attrdict, raw=False, use_list=False)

        async with anyio.create_task_group() as tg:
            self.tg = tg
            await self.send({'seq': 0, 'version': _version_tuple,
                'node':self.server.node.name, 'local':self.server.node.tick})

            while True:
                for msg in unpacker:
                    try:
                        logger.debug("IN %d: %s", self._client_nr, msg)
                        seq = msg.seq
                        q = self.inqueue.get(seq, None)
                        if q is not None:
                            await q.put(msg)
                        elif msg.get('state','') == 'start':
                            q = anyio.create_queue(100)
                            self.inqueue[seq] = q
                            await self.tg.spawn(self.process, q, True)
                        else:
                            if self.seq >= seq:
                                raise ClientError("Sequence numbers are not monotonic: %d < %d" % (self.seq, msg.seq))
                            self.seq = seq
                            await self.tg.spawn(self.process, msg)
                    except Exception as exc:
                        if not isinstance(exc, ClientError):
                            logger.exception("Client error on %s", repr(msg))
                        await self.send({'error': str(exc)})

                try:
                    buf = await self.stream.receive_some(4096)
                except (ConnectionResetError, ClosedResourceError):
                    return  # closed/reset/whatever
                if len(buf) == 0:  # Connection was closed.
                    return # done
                unpacker.feed(buf)


class Server:
    serf = None

    def __init__(self, name: str, root: Entry, host: str, port: int):
        self.root = root
        self.host = host
        self.port = port
        self.node = Node(name, 1)

        self._nodes = {self.node.name: self.node}
        self._evt_lock = anyio.create_lock()

    @asynccontextmanager
    async def next_event(self):
        async with self._evt_lock:
            self.node.tick += 1
            yield NodeEvent(self.node)

    async def watcher(self):
        async with Watcher(self.root) as watcher:
            async for msg in watcher:
                if msg.event.node != self.node:
                    continue
                p = msg.serialize(nchain=9)
                p['action'] = 'update'
                p = _packer(p)
                await self.serf.event('distkv.update', p, coalesce=False)

    async def _process(self, msg):
        try:
            fn = getattr(self, 'cmd_' + str(msg.action))
        except AttributeError:
            raise ClientError("Command not recognized: " + repr(msg.action))
        else:
            return await fn(msg)

    async def cmd_update(self, msg):
        msg = UpdateEvent.unserialize(self.root, msg)
        await msg.entry.apply(msg)

    async def monitor(self):
        async with self.serf.stream('user:distkv.update') as stream:
            async for resp in stream:
                msg = msgpack.unpackb(resp.payload, object_pairs_hook=attrdict, raw=False, use_list=False)

                evt = await self._process(msg)

    async def serve(self, cfg={}):
        async with serf_client(**cfg.serf) as serf:
            self.serf = serf
            self.spawn = serf.spawn

            await serf.spawn(self.monitor)
            await serf.spawn(self.watcher)
            cfg_s = cfg.server.copy()
            cfg_s.pop('host', None)
            cfg_s['port'] = self.port
            async with await anyio.create_tcp_server(**cfg_s) as server:
                async for client in server.accept_connections():
                    await serf.spawn(self._connect, client)

    async def _connect(self, stream):
        c = ServerClient(server=self, root=self.root, stream=stream)
        await c.run()




# msgpack and co.

import msgpack
from functools import partial

from .util import attrdict

__all__ = ['packer','unpacker', 'stream_unpacker']

# single message packer
packer = msgpack.Packer(strict_types=False, use_bin_type=True).pack

# single message unpacker
unpacker = partial(msgpack.unpackb, object_pairs_hook=attrdict, raw=False, use_list=False)

# Data stream unpacker factory (client)
stream_unpacker = lambda: msgpack.Unpacker(object_pairs_hook=attrdict, raw=False, use_list=False)




# command line interface

import sys
import trio_click as click
from pprint import pprint

from .util import attrdict, combine_dict, PathLongener
from .client import open_client, StreamReply
from .default import CFG, PORT
from .server import Server
from .model import Entry

import trio_click as click

import logging
logger = logging.getLogger(__name__)

class _NotGiven:
    pass
_NotGiven = _NotGiven()

def cmd():
    try:
        main(standalone_mode=False)
    except BaseException as exc:
        raise
        print(exc)
        sys.exit(1)

@click.group()
@click.option("-v","--verbose", count=True, help="Enable debugging. Use twice for more verbosity.")
@click.option("-q","--quiet", count=True, help="Disable debugging. Opposite of '--verbose'.")
@click.option("-c","--cfg", type=click.File('r'), default=None)
@click.pass_context
async def main(ctx, verbose, quiet, cfg):
    ctx.ensure_object(attrdict)
    ctx.obj.debug = verbose - quiet
    logging.basicConfig(level=logging.DEBUG if verbose>2 else
                              logging.INFO if verbose>1 else
                              logging.WARNING if verbose>0 else
                              logging.ERROR)
    if cfg:
        logger.debug("Loading %s",cfg)
        import yaml
        ctx.obj.cfg = combine_dict(yaml.safe_load(cfg), CFG)
        cfg.close()
    else:
        ctx.obj.cfg = CFG

@main.command()
@click.option("-h","--host", default=None, help="Address to bind to. Default: %s" % (CFG.server.host))
@click.option("-p","--port", type=int, default=None, help="Port to bind to. Default: %d" % (CFG.server.port,))
@click.option("-l","--load", type=click.File('rb'), default=None, help="Event log to preload.")
@click.option("-s","--save", type=click.Path('wb'), default=None, help="Event log to write.")
@click.argument("name", nargs=1)
@click.pass_context
async def run(ctx, name, host, port, load, save):
    obj = ctx.obj
    print("Start run.")
    if host is None:
        host = obj.cfg.server.host
    if port is None:
        port = obj.cfg.server.port

    obj.root = Entry("ROOT", None)
    s = Server(name, obj.root, host, port)
    if load is None:
        load = obj.cfg['state']
        await s.load(load)
    if save is not None:
        if save == '':
            save = None
        obj.cfg['state'] = save
    await s.serve(obj.cfg)


@main.group()
@click.option("-h","--host", default=None, help="Host to use. Default: %s" % (CFG.server.host,))
@click.option("-p","--port", type=int, default=None, help="Port to use. Default: %d" % (CFG.server.port,))
@click.pass_context
async def client(ctx,host,port):
    obj = ctx.obj
    if host is None:
        host = obj.cfg.server.host
    if port is None:
        port = obj.cfg.server.port
    obj.client = await ctx.enter_async_context(open_client(host, port))
    logger.debug("Connected.")


@client.command()
@click.option("-c", "--chain", default=0, help="Length of change list to return. Default: 0")
@click.option("-y", "--yaml", is_flag=True, help="Print as YAML. Default: Python.")
@click.option("-d", "--as-dict", default=None, help="YAML: structure as dictionary. The argument is the key to use for values. Default: return as list")
@click.option("-v", "--verbose", is_flag=True, help="Print the complete result. Default: just the value")
@click.option("-r", "--recursive", is_flag=True, help="Read a complete subtree")
@click.argument("path", nargs=-1)
@click.pass_context
async def get(ctx, path, chain, yaml, verbose, recursive, as_dict):
    """Read a DistKV value"""
    obj = ctx.obj
    if verbose and yaml:
        raise click.UsageError("'verbose' and 'yaml' are mutually exclusive")
    if recursive:
        if verbose and yaml:
            raise click.UsageError("'verbose' does not work in recursive YAML mode")
        res = await obj.client.request(action="get_tree", path=path, iter=True, nchain=chain)
        pl = PathLongener(path)
        y = {} if as_dict is not None else []
        async for r in res:
            pl(r)
            if yaml:
                if as_dict is not None:
                    yy = y
                    for p in r.path:
                        yy = yy.setdefault(p,{})
                        yy[as_dict] = r.value
                else:
                    yr = {'path': r.path, 'value': r.value}
                    if 'chain' in r:
                        yr['chain'] = r.chain
                    y.append(yr)
            else:
                if verbose:
                    pprint(r)
                else:
                    print("%s: %s" % (' '.join(r.path), repr(r.value)))
        if yaml:
            import yaml
            print(yaml.safe_dump(y))
        return
    res = await obj.client.request(action="get_value", path=path, iter=False, nchain=chain)
    if not verbose:
        res = res.value
    if yaml:
        import yaml
        print(yaml.safe_dump(res))
    else:
        pprint(res)


@client.command()
@click.option("-v", "--value", help="Value to set. Mandatory.")
@click.option("-e", "--eval", is_flag=True, help="The value shall be evaluated.")
@click.option("-c", "--chain", default=0, help="Length of change list to return. Default: 0")
@click.option("-p", "--prev", default=_NotGiven, help="Previous value. Deprecated; use 'last'")
@click.option("-l", "--last", nargs=2, help="Previous change entry (node serial)")
@click.option("-y", "--yaml", is_flag=True, help="Print result as YAML. Default: Python.")
@click.argument("path", nargs=-1)
@click.pass_context
async def set(ctx, path, value, eval, chain, prev, last, yaml):
    """Set a DistKV value"""
    obj = ctx.obj
    if eval:
        value = __builtins__['eval'](value)
    args = {}
    if prev is not _NotGiven:
        if eval:
            prev = __builtins__['eval'](prev)
        args['prev'] = prev
    if last:
        if last[1] == '-':
            args['chain'] = None
        else:
            args['chain'] = {'node': last[0], 'tick': int(last[1])}

    res = await obj.client.request(action="set_value", value=value, path=path, iter=False, nchain=chain, **args)
    if yaml:
        import yaml
        print(yaml.safe_dump(res))
    elif chain:
        pprint(res)


@client.command()
@click.argument("path", nargs=-1)
@click.option("-c", "--chain", default=0, help="Length of change list to return. Default: 0")
@click.option("-r", "--recursive", is_flag=True, help="Delete a complete subtree")
@click.pass_context
async def delete(ctx, path, chain, recursive):
    obj = ctx.obj
    res = await obj.client.request(action="delete_tree" if recursive else "delete_value", path=path, nchain=chain)
    if isinstance(res, StreamReply):
        pl = PathLongener(path)
        async for r in res:
            pl(r)
            pprint(r)
    else:
        pprint(res)


@client.command()
@click.option("-c", "--chain", default=0, help="Length of change list to return. Default: 0")
@click.option("-y", "--yaml", is_flag=True, help="Print as YAML. Default: Python.")
@click.argument("path", nargs=-1)
@click.pass_context
async def watch(ctx, path, chain, yaml):
    """Watch a DistKV subtree"""
    if yaml:
        import yaml
    obj = ctx.obj
    res = await obj.client.request(action="watch", path=path, iter=True, nchain=chain)
    pl = PathLongener(path)
    async for r in res:
        pl(r)
        del r['seq']
        if yaml:
            print(yaml.safe_dump(r))
        else:
            pprint(r)

@client.command()
@click.option("-l", "--local", is_flag=True, help="Load locally, don't broadcast")
@click.option("-f", "--force", is_flag=True, help="Overwrite existing values")
@click.option("-i", "--infile", type=click.File('rb'), help="Print as YAML. Default: Python.")
@click.argument("path", nargs=-1)
@click.pass_context
async def update(ctx, path, infile, loca, force):
    """Send a list of updates to a DistKV subtree"""
    if local and force:
        raise click.UsageError("'local' and 'force' are mutually exclusive")

    obj = ctx.obj
    ps = PathShortener()
    async with MsgReader() as reader:
        with obj.client.stream(action="update", path=path, iter=False, force=force, local=local) as sender:
            async for r in res:
                ps(r)
                await sender(r)

    print(sender.result)
